"""
High-level interface for data access using Table Access Protocol (TAP)
"""


import os
import secrets
import shutil
import tempfile
import time
from functools import wraps
from getpass import getpass
from inspect import getfullargspec
from io import BytesIO
from pathlib import Path
from typing import Callable, Dict, List, Literal, Sequence, Union
from urllib.parse import quote, urlencode

import numpy as np
import pandas as pd
import requests
from astropy import units as u
from astropy.table import Table
from astropy.units import Quantity
from bs4 import BeautifulSoup
from dl import authClient as ac
from dl import queryClient as qc

from pylegs.config import configs
from pylegs.io import (PathLike, PathOrFile, TableLike, _create_parents,
                       _prepare_path, concat_tables, download_file,
                       parallel_function_executor, read_table, write_table)
from pylegs.utils import SingletonMeta, guess_coords_columns, sanitize_quantity

__all__ = [
  'dl_crossmatch', 'sync_query', 'async_query', 'batch_sync_query', 
  'batch_async_query'
]

DEFAULT_CATALOG_COLUMNS = [
  'ra', 'dec', 'fracflux_r', 'fracin_r', 'fracmasked_r',
  'mag_g', 'mag_i', 'mag_r', 'mag_z',
  'nea_g', 'nea_i', 'nea_r', 'nea_z', 'nest4096', 'ngood_g', 'ngood_i', 
  'ngood_r', 'ngood_z', 'nobs_g', 'nobs_i', 'nobs_r', 'nobs_z', 'ring256', 
  'sersic', 'shape_e1', 'shape_e2', 'shape_r', 'snr_g', 'snr_i', 'snr_r', 
  'snr_z', 'type'
]



class DataLabAuth(metaclass=SingletonMeta):
  def __init__(self):
    self._token: str | None = None
    self._username: str | None = None
    self._password: str | None = None
  
  def login(self, username: str, password: str) -> str:
    self._username = username
    self._password = password
    return self.token
  
  @property
  def token(self) -> str:
    if self._token is None or not ac.isTokenLoggedIn(self._token):
      username = self._username or configs.DATALAB_USERNAME or os.environ.get('DATALAB_USERNAME')
      password = self._password or configs.DATALAB_PASSWORD or os.environ.get('DATALAB_PASSWORD')
      if username is None:
        username = input('\nDatalab Username: ')
      if password is None:
        password = getpass('Datalab Password: ')
      if ac.isValidUser(username):
        self._token = ac.login(username, password)
    return self._token



def dl_auth(func):
  @wraps(func)
  def wrapper(*args, **kwargs):
    username = kwargs.get('username')
    password = kwargs.get('password')
    DataLabAuth().login(username, password)
    return func(*args, **kwargs)
  return wrapper



def _dl_crossmatch(
  table: TableLike = None,
  table_cols: Sequence[str] = None,
  catalog: str = 'ls_dr10.tractor',
  catalog_cols: Sequence[str] = None,
  output_path: PathOrFile = None,
  join: Literal['left', 'right', 'outer', 'inner', 'cross'] = 'left',
  radius: float | str | Quantity = '1 arcsec',
  tab_ra: str = None,
  tab_dec: str = None,
  cat_ra: str = 'ra',
  cat_dec: str = 'dec',
):
  token = DataLabAuth().token
    
  df = read_table(table)
  tab_ra, tab_dec = guess_coords_columns(df, tab_ra, tab_dec)
  df = df[[tab_ra, tab_dec]].copy()
  df['_pylegs_key'] = range(len(df))
  df = df.rename(columns={tab_ra: 'ra', tab_dec: 'dec'})
  
  radius = sanitize_quantity(radius, u.deg, True)
  
  mydb_table_name = f'pylegs_input_{secrets.token_hex(4)}'
  qc.mydb_import(token, mydb_table_name, df)
  
  sql_template = """
    SELECT local._pylegs_key, {remote_columns},
    (q3c_dist(local.ra, local.dec, remote.{cat_ra}, remote.{cat_dec})*3600.0) as dist_arcsec 
    FROM mydb://{mydb_table_name} AS local 
    INNER JOIN LATERAL (
      SELECT cat.* FROM {catalog} AS cat
      WHERE q3c_join(local.ra, local.dec, cat.{cat_ra}, cat.{cat_dec}, {sep_deg}) 
      ORDER BY q3c_dist(local.ra, local.dec, cat.{cat_ra}, cat.{cat_dec}) ASC LIMIT 1
    ) as remote ON true
  """
  sql = sql_template.format(
    remote_columns=', '.join([f'remote.{c}' for c in catalog_cols]),
    mydb_table_name=mydb_table_name,
    catalog=catalog,
    sep_deg=radius.value,
    cat_ra=cat_ra,
    cat_dec=cat_dec,
  )

  match_df = qc.query(
    token=token,
    sql=sql,
    fmt='pandas',
    out=None,
    drop=True,
    async_=False,
    timeout=600,
  )
  
  qc.mydb_drop(mydb_table_name, token=token)
  
  df = read_table(table, columns=table_cols)
  df['_pylegs_key'] = range(len(df))
  result_df = df.set_index('_pylegs_key').join(
    match_df.set_index('_pylegs_key'), how=join
  )
  
  if output_path is not None:
    write_table(result_df, output_path)
  
  return result_df



@dl_auth
def dl_crossmatch(
  table: TableLike = None,
  table_cols: Sequence[str] = None,
  catalog: str = 'ls_dr10.tractor',
  catalog_cols: Sequence[str] = None,
  output_path: PathOrFile = None,
  join: Literal['left', 'inner'] = 'inner',
  radius: float | str | Quantity = '1 arcsec',
  tab_ra: str = None,
  tab_dec: str = None,
  cat_ra: str = 'ra',
  cat_dec: str = 'dec',
  cache_dir: PathLike | None = None,
  workers: int = 5,
  overwrite: bool = False,
  **kwargs
):
  if output_path and isinstance(output_path, PathLike):
    output_path = Path(output_path)
    if output_path.exists() and not overwrite:
      return

  chunk_rows = 100_000
  df = read_table(table)
  n_splits = int(np.ceil(len(df) / chunk_rows))
  
  func_args = {
    'table': table,
    'table_cols': table_cols,
    'catalog': catalog,
    'catalog_cols': catalog_cols,
    'output_path': output_path,
    'join': join,
    'radius': radius,
    'tab_ra': tab_ra,
    'tab_dec': tab_dec,
    'cat_ra': cat_ra,
    'cat_dec': cat_dec,
  }
  
  if n_splits < 2:
    return _dl_crossmatch(**func_args)
  else:
    if cache_dir is None:
      cache_dir = Path(tempfile.gettempdir()) / f'pylegs_xmatch_{secrets.token_hex(4)}'
      cache_dir.mkdir(parents=True, exist_ok=True)
    else:
      cache_dir = Path(cache_dir)
      cache_dir.mkdir(parents=True, exist_ok=True)
    
    params = [
      {
        **func_args, 
        'table': df.iloc[i*chunk_rows : (i+1)*chunk_rows].copy(),
        'output_path': cache_dir / f'{i}.parquet'
      } 
      for i in range(n_splits)
    ]
    
    parallel_function_executor(
      func=_dl_crossmatch,
      params=params,
      workers=workers,
      progress=True,
      ignore_error=False,
      unit='query',
    )
    
    df = concat_tables(sorted([str(p.absolute()) for p in cache_dir.glob('*')]))
    shutil.rmtree(cache_dir)
    
    if output_path:
      write_table(df, output_path)
        
    return df



@dl_auth
def print_mydb_tables(**kwargs):
  token = DataLabAuth().token
  ls = qc.list(token, '')
  print(ls)



@dl_auth
def drop_mydb_pylegs(**kwargs):
  token = DataLabAuth().token
  tables = qc.list(token, '')
  tables = tables.split('\n')
  for t in tables:
    if t.startswith('pylegs'):
      qc.mydb_drop(token, t)


@dl_auth
def drop_mydb_table(table: str, **kwargs):
  token = DataLabAuth().token
  qc.mydb_drop(token, table)



def sync_query(
  query: str, 
  save_path: PathOrFile = None,
  overwrite: bool = True,
  http_client: requests.Session = configs.HTTP_CLIENT,
  dtype: Literal['bytes', 'pandas', 'astropy'] = 'bytes'
) -> bytes | pd.DataFrame | Table:
  params = {
    'request': 'doQuery',
    'version': 1.0,
    'lang': 'ADQL',
    'phase': 'run',
    'format': 'csv',
    'query': query
  }
  save_path = _prepare_path(save_path)
  _create_parents(save_path)
  
  # req_url = configs.TAP_SYNC_BASE_URL + '?' + urlencode(params)
  table_bytes = None
  attempt = 0
  max_attempts = 5
  while table_bytes is None and attempt < max_attempts:
    table_bytes = download_file(
      url=configs.TAP_SYNC_BASE_URL,#req_url, 
      save_path=save_path, 
      overwrite=overwrite,
      query=params,
      http_client=http_client,
    )
    attempt += 1
  
  if dtype == 'bytes':
    return table_bytes
  if dtype in ('pandas', 'astropy'):
    return read_table(BytesIO(table_bytes), fmt='csv', dtype=dtype)


def async_query(
  query: str, 
  save_path: PathOrFile = None,
  overwrite: bool = True,
  http_client: requests.Session = configs.HTTP_CLIENT,
  delay: int = 5,
  dtype: Literal['bytes', 'pandas', 'astropy'] = 'bytes'
) -> bytes | pd.DataFrame | Table:
  params = {
    'request': 'doQuery',
    'version': 1.0,
    'lang': 'ADQL',
    'phase': 'run',
    'format': 'csv',
    'query': query
  }
  save_path = _prepare_path(save_path)
  _create_parents(save_path)
  
  table_bytes = None
  attempt = 0
  max_attempts = 5
  
  while table_bytes is None and attempt < max_attempts:
    resp = http_client.post(
      url=configs.TAP_ASYNC_BASE_URL,
      data=params,
    )
    soup = BeautifulSoup(resp.text, 'xml')
    
    job_id = soup.find('uws:jobId').text
    job_phase = soup.find('uws:phase').text
    table_bytes = None
    
    while job_phase == 'PENDING':
      time.sleep(delay)
      resp = http_client.get(configs.TAP_ASYNC_BASE_URL + f'/{job_id}')
      soup = BeautifulSoup(resp.text, 'xml')
      job_phase = soup.find('uws:phase').text
    
    if job_phase == 'COMPLETED':
      table_url = soup.find('#result').attrs['xlink:href']
      table_bytes = download_file(
        url=table_url, 
        save_path=save_path, 
        overwrite=overwrite, 
        http_client=http_client
      )
    attempt += 1
  
  if dtype == 'bytes':
    return table_bytes
  if dtype in ('pandas', 'astropy'):
    return read_table(BytesIO(table_bytes), fmt='csv', dtype=dtype)
  
  
def _batch_query(
  func: Callable,
  queries: Sequence[str],
  save_paths: Sequence[str | Path],
  func_args: Dict[str, None],
  workers: int = 3,
  concat: bool = False,
  partial_paths: Sequence[str | Path] | None = None,
):
  save_paths_aux = save_paths
  if concat:
    if partial_paths is None:
      tmp_folder = Path(tempfile.gettempdir()) / f'pylegs_tap_{secrets.token_hex(4)}'
      tmp_folder.mkdir(parents=True, exist_ok=True)
      save_paths_aux = [tmp_folder / f'{i}.csv' for i in range(len(queries))]
    else:
      save_paths_aux = [Path(i) for i in partial_paths]
      save_paths_aux[0].parent.mkdir(parents=True, exist_ok=True)
    
  params = [
    {
      'query': _query,
      'save_path': _save_path,
      **func_args,
    }
    for _query, _save_path in zip(queries, save_paths_aux)
  ]
  
  try:
    parallel_function_executor(
      func,
      params=params,
      workers=workers,
      unit='query',
      ignore_error=False,
    )

    if concat:
      combined_df = concat_tables([p for p in save_paths_aux if p.exists()])
      write_table(combined_df, save_paths)
      if partial_paths is None:
        shutil.rmtree(tmp_folder)
  except Exception as e:
    if concat and partial_paths is None:
      shutil.rmtree(tmp_folder)
    raise e


def batch_sync_query(
  queries: str, 
  save_paths: PathOrFile = None,
  overwrite: bool = True,
  concat: bool = False,
  workers: int = 3,
  http_client: requests.Session = configs.HTTP_CLIENT,
  partial_paths: Sequence[str | Path] | None = None,
  overwrite_partials: bool = False,
) -> bytes | pd.DataFrame | Table:
  args = {
    'overwrite': overwrite_partials if concat else overwrite,
    'http_client': http_client,
  }
  _batch_query(
    func=sync_query, 
    queries=queries, 
    save_paths=save_paths, 
    func_args=args, 
    workers=workers, 
    concat=concat,
    partial_paths=partial_paths,
  )
  
  
def batch_async_query(
  queries: str, 
  save_paths: PathOrFile = None,
  overwrite: bool = True,
  concat: bool = False,
  workers: int = 3,
  http_client: requests.Session = configs.HTTP_CLIENT,
  delay: int = 5,
  partial_paths: Sequence[str | Path] | None = None,
  overwrite_partials: bool = False,
) -> bytes | pd.DataFrame | Table:
  args = {
    'overwrite': overwrite,
    'http_client': http_client,
    'delay': delay,
  }
  _batch_query(
    func=async_query, 
    queries=queries, 
    save_paths=save_paths, 
    func_args=args, 
    workers=workers, 
    concat=concat,
    partial_paths=partial_paths,
  )



def download_catalog(
  save_path: PathOrFile,
  columns: Sequence[str] | None = None,
  ra_min: float | Quantity = 0,
  ra_max: float | Quantity = 360,
  delta_ra: float | Quantity = 10 * u.arcmin,
  table: str = 'ls_dr10.tractor',
  exclude_types: Sequence[str] | None = ['PSF'],
  magr_min: float = 10,
  magr_max: float = 21,
  dec_min: float | Quantity = -90,
  dec_max: float | Quantity = 90, 
  brick_primary: bool | None = True,
  overwrite: bool = False,
  workers: int = 6,
  tmp_folder: str | Path | None = None,
  overwrite_tmp: bool = False,
):
  """
  PSF, REX, DEV, EXP, SER  
  """
  if columns is None:
    columns = DEFAULT_CATALOG_COLUMNS
    
  filters = ''
  if brick_primary is not None:
    filters += f'AND brick_primary = {int(brick_primary)} '
  if exclude_types is not None:
    for t in exclude_types:
      filters += f"AND type != '{t.upper()}' "
  
  template = """
  SELECT {cols}
  FROM {table}
  WHERE ra BETWEEN {ra_min:.12f} AND {ra_max:.12f}
  AND dec BETWEEN {dec_min:.8f} AND {dec_max:.8f}
  AND mag_r BETWEEN {magr_min:.7f} AND {magr_max:.7f} 
  {filters}
  """.strip()
  
  print('Summary:')
  ra_min = sanitize_quantity(ra_min, u.deg, convert=True)
  ra_max = sanitize_quantity(ra_max, u.deg, convert=True)
  print(f'RA range: [{ra_min}, {ra_max}]')
  ra_min, ra_max = ra_min.value, ra_max.value
  
  dec_min = sanitize_quantity(dec_min, u.deg, convert=True)
  dec_max = sanitize_quantity(dec_max, u.deg, convert=True)
  print(f'DEC range: [{dec_min}, {dec_max}]')
  dec_min, dec_max = dec_min.value, dec_max.value

  delta_ra = sanitize_quantity(delta_ra, u.deg, convert=True)
  print(f'RA step: {delta_ra}')
  delta_ra = delta_ra.value
  
  queries = [
    template.format(
      cols=', '.join(columns), 
      table=table, 
      magr_min=magr_min, 
      magr_max=magr_max, 
      filters=filters,
      dec_min=dec_min,
      dec_max=dec_max,
      ra_min=_ra_min,
      ra_max=_ra_min + delta_ra,
    )
    for _ra_min in np.arange(ra_min, ra_max, delta_ra)
  ]
  
  tmp_paths = None
  if tmp_folder is not None:
    tmp_paths = [
      Path(tmp_folder) / f'query_{_ra:.6f}-{(_ra+delta_ra):.6f}.csv' 
      for _ra in np.arange(ra_min, ra_max, delta_ra)
    ]
  
  print('\nExample query:')
  print(queries[0])
  print(f'\nNumber of queries: {len(queries)}\n')
  
  return batch_sync_query(
    queries=queries, 
    save_paths=save_path, 
    overwrite=overwrite, 
    concat=True, 
    workers=workers,
    partial_paths=tmp_paths,
    overwrite_partials=overwrite_tmp,
  )




if __name__ == '__main__':
  # sync_query(
  #   'select top 10 psfdepth_g, psfdepth_r from ls_dr9.tractor where ra between 230.2939-0.0013 and 230.2939+0.0013 and dec between 29.7714-0.0013 and 29.7714+0.0013',
  #   url='https://datalab.noirlab.edu/tap/sync'
  # )
  # sync_query('select top 10 psfdepth_g, psfdepth_r from ls_dr9.tractor where ra between 230.2939-0.0013 and 230.2939+0.0013 and dec between 29.7714-0.0013 and 29.7714+0.0013')
  download_catalog(ra_min=120, ra_max=120*u.deg+2*u.arcmin, save_path='test.parquet', delta_ra=1*u.arcmin, magr_min=15, magr_max=16, overwrite=True)


import bz2
import concurrent.futures
import re
from io import BufferedIOBase, BytesIO, RawIOBase, StringIO, TextIOBase
from pathlib import Path
from typing import Any, Callable, Dict, Literal, Sequence

import pandas as pd
import requests
from astropy.io import fits, votable
from astropy.table import Table
from tqdm import tqdm

__all__ = [
  'download_file', 'parallel_function_executor', 'compress_fits_image',
  'read_table', 'write_table'
]


PathLike = str | Path
FileLike = BytesIO
PathOrFile = PathLike | FileLike
TableLike = pd.DataFrame | Table


def _prepare_path(path: PathOrFile) -> PathOrFile:
  if isinstance(path, str):
    return Path(path)
  return path


def _path_exists(path: PathOrFile) -> bool:
  if isinstance(path, Path):
    return path.exists()
  if isinstance(path, str):
    return Path(path).exists()
  return False


def _create_parents(path: PathOrFile):
  if isinstance(path, (str, Path)):
    Path(path).parent.mkdir(parents=True, exist_ok=True)
    
    
def _write_bytes(path: PathOrFile, b: bytes):
  if isinstance(path, (str, Path)):
    with open(path, 'wb') as f:
      f.write(b)
  if isinstance(path, BytesIO):
    path.write(b)
    
    
def _is_folder(path: PathOrFile) -> bool:
  if isinstance(path, (str, Path)):
    return Path(path).is_dir()
  return False


def download_file(
  url: str,
  save_path: PathOrFile = None,
  query: Dict[str, Any] = None,
  overwrite: bool = False,
  http_client: requests.Session = None,
  extract: bool = False,
) -> bytes | None:
  """
  Download a file from a URL and save it to a specified path.

  Parameters
  ----------
  url : str
    The URL from which to download the file.
  save_path : PathOrFile
    The path where the downloaded file will be saved. This can be a string or 
    a `pathlib.Path` object.
  query : Dict[str, Any], optional
    A dictionary of query parameters to include in the request. Defaults to ``None``.
  overwrite : bool, optional
    If True, overwrite the existing file at `save_path`. Defaults to ``False``.
  http_client : requests.Session, optional
    The HTTP client to use for making the request. If None, the `requests` 
    library is used. Defaults to ``None``.
  extract : bool, optional
    If ``True``, decompress the downloaded file if it is in ``bz2`` format. 
    Defaults to ``False``.

  Returns
  -------
  bytes or None
    The bytes of the downloaded (and possibly decompressed) file, or None if 
    the file already exists and `overwrite` is False.

  Notes
  -----
  - If ``overwrite`` is ``False`` and the file already exists at ``save_path``, 
    the function returns ``None`` without downloading the file.
  - If ``extract`` is ``True``, the function assumes the file is compressed 
    with bz2 and decompresses it before saving.

  Raises
  ------
  Exception
    If the HTTP request fails or if there is an error in writing the file.

  Examples
  --------
  >>> download_file('http://example.com/file.txt', 'file.txt')
  b'file content'
  
  >>> download_file('http://example.com/file.bz2', 'file.txt', extract=True)
  b'decompressed content'
  """
  save_path = _prepare_path(save_path)

  if _path_exists(save_path) and not overwrite:
    return None

  _create_parents(save_path)

  http_client = http_client or requests

  r = http_client.get(url, params=query, allow_redirects=True)

  file_bytes = None
  if r.status_code == 200:
    if extract:
      file_bytes = bz2.decompress(r.content)
    else:
      file_bytes = r.content
    
    if save_path is not None:
      _write_bytes(save_path, file_bytes)

  return file_bytes


def compress_fits_image(
  file: PathOrFile,
  save_path: str | Path = None,
  overwrite: bool = True,
  compress_type: str = 'HCOMPRESS_1',
  hcomp_scale: int = 3,
  quantize_level: int = 10,
  quantize_method: int = -1,
  ext: int = 0,
  random_seed: int = 42,
):
  """
  Compress a FITS image and optionally save the compressed image to a file.

  Parameters
  ----------
  file : PathOrFile
    The path to the FITS file to be compressed. This can be a string or a 
    `pathlib.Path` object.
  save_path : str or Path, optional
    The path where the compressed FITS image will be saved. If None, 
    the compressed image is not saved.
    Defaults to None.
  overwrite : bool, optional
    If ``True``, overwrite the existing file at ``save_path``. Defaults to ``True``.
  compress_type : str, optional
    The compression algorithm to use. Supported values include 
    ``'HCOMPRESS_1'``, ``'RICE_1'``, ``'GZIP_1'``, ``'PLIO_1'``.
    Defaults to 'HCOMPRESS_1'.
  hcomp_scale : int, optional
    The scale parameter for HCOMPRESS. Defaults to 3.
  quantize_level : int, optional
    The quantization level for compression. Defaults to 10.
  quantize_method : int, optional
    The quantization method to use. Defaults to -1.
  ext : int, optional
    The ``ext`` of the FITS file to compress. Defaults to 0.
  random_seed : int, optional
    The seed for random number generation used in dithering. Defaults to 42.

  Returns
  -------
  fits.CompImageHDU
    The compressed image HDU.

  Raises
  ------
  ValueError
    If the specified FITS ``ext`` is out of range.
  
  Notes
  -----
  If ``save_path`` is provided and ``overwrite`` is True, the compressed 
  image will be saved to the specified path. Otherwise, the function returns 
  the compressed image HDU.

  Examples
  --------
  >>> compress_fits_image('input.fits', save_path='compressed.fits', compress_type='RICE_1')
  <astropy.io.fits.hdu.compressed.CompImageHDU object at 0x...>
  """
  hdul = fits.open(file)

  if ext >= len(hdul):
    raise ValueError(f'Trying to access ext {ext}, max ext is {len(hdul)-1}')

  if _path_exists(save_path) and not overwrite:
    return None

  comp = None
  comp = fits.CompImageHDU(
    data=hdul[ext].data,
    header=hdul[ext].header,
    compression_type=compress_type,
    hcomp_scale=hcomp_scale,
    quantize_level=quantize_level,
    quantize_method=quantize_method,
    dither_seed=random_seed,
  )
  if save_path:
    comp.writeto(save_path, overwrite=overwrite)
  
  return comp


def read_table(
  path: TableLike | PathOrFile,
  fmt: str | None = None,
  columns: Sequence[str] | None = None,
  low_memory: bool = False,
  comment: str | None = None,
  na_values: Sequence[str] | Dict[str, Sequence[str]] = None,
  keep_default_na: bool = True,
  na_filter: bool = True,
  header: Literal['infer'] | int | Sequence[int] = 'infer',
  col_names: Sequence[str] | None = None,
  dtype: Literal['pandas', 'astropy'] = 'pandas',
) -> pd.DataFrame:
  """
  This function tries to detect the table type comparing the file extension and
  returns a pandas dataframe of the loaded table.
  
  Supported table types:
  
    =============== ===========================
    Table Type      Extensions
    =============== ===========================
    Fits            .fit, .fits, .fz
    Votable         .vo, .vot, .votable, .xml
    ASCII           .csv, .tsv, .dat
    Heasarc         .tdat
    Arrow           .parquet, .feather
    =============== ===========================

  Parameters
  ----------
  path : str or Path
    Path to the table to be read.
  
  fmt : str or None
    Specify the file format manually to avoid inference by file extension. This
    parameter can be used to force a specific parser for the given file.
  
  columns : sequence of str or None
    If specified, only the column names in list will be loaded. Can be used to
    reduce memory usage.
  
  low_memory : bool
    Internally process the file in chunks, resulting in lower memory use while 
    parsing, but possibly mixed type inference. To ensure no mixed types either 
    set False, or specify the type with the dtype parameter. Note that the 
    entire file is read into a single DataFrame regardless, use the chunksize 
    or iterator parameter to return the data in chunks. (Only valid with C parser).
    
    .. note::
      Used only for ASCII tables, ignored by other types of tables.
  
  comment : str or None
    Character indicating that the remainder of line should not be parsed. 
    If found at the beginning of a line, the line will be ignored altogether. 
    This parameter must be a single character. Like empty lines 
    (as long as ``skip_blank_lines=True``), fully commented lines are ignored 
    by the parameter header but not by skiprows. For example, if ``comment='#'``, 
    parsing ``#empty\\na,b,c\\n1,2,3`` with ``header=0`` will result in 
    ``'a,b,c'`` being treated as the header.
    
    .. note::
      Used only for ASCII tables, ignored by other types of tables.
  
  na_values: hashable, iterable of hashable or dict of `HashableIterable`
    Additional strings to recognize as ``NA``/``NaN``. If ``dict`` passed, specific 
    per-column ``NA`` values. By default the following values are interpreted 
    as `NaN`: “ “, “#N/A”, “#N/A N/A”, “#NA”, “-1.#IND”, “-1.#QNAN”, “-NaN”, 
    “-nan”, “1.#IND”, “1.#QNAN”, “<NA>”, “N/A”, “NA”, “NULL”, “NaN”, “None”, 
    “n/a”, “nan”, “null “.
    
    .. note::
      Used only for ASCII tables, ignored by other types of tables.
  
  keep_default_na : bool 
    Whether or not to include the default ``NaN`` values when parsing the data. 
    Depending on whether ``na_values`` is passed in, the behavior is as follows:

    - If ``keep_default_na`` is ``True``, and ``na_values`` are specified, 
      `na_values` is appended to the default NaN values used for parsing.
    - If ``keep_default_na`` is ``True``, and ``na_values`` are not specified, only the 
      default ``NaN`` values are used for parsing.
    - If ``keep_default_na`` is ``False``, and ``na_values`` are specified, only 
      the ``NaN`` values specified na_values are used for parsing.
    - If ``keep_default_na`` is ``False``, and ``na_values`` are not specified, 
      no strings will be parsed as ``NaN``.

    Note that if ``na_filter`` is passed in as ``False``, the ``keep_default_na`` and 
    ``na_values`` parameters will be ignored.
    
    .. note::
      Used only for ASCII tables, ignored by other types of tables.
  
  na_filter : bool
    Detect missing value markers (empty strings and the value of ``na_values``). 
    In data without any ``NA`` values, passing ``na_filter=False`` can improve the 
    performance of reading a large file.
    
    .. note::
      Used only for ASCII tables, ignored by other types of tables.
      
  header : 'infer' or int or sequence of int
    Row number(s) containing column labels and marking the start of the data 
    (zero-indexed). Default behavior is to infer the column names: if no ``names``
    are passed the behavior is identical to ``header=0`` and column names are 
    inferred from the first line of the file, if column names are passed 
    explicitly to ``names`` then the behavior is identical to ``header=None``. 
    Explicitly pass ``header=0`` to be able to replace existing names. The 
    header can be a list of integers that specify row locations for a 
    `pandas.MultiIndex` on the columns e.g. ``[0, 1, 3]``. Intervening rows 
    that are not specified will be skipped (e.g. 2 in this example is skipped). 
    Note that this parameter ignores commented lines and empty lines if 
    ``skip_blank_lines=True``, so ``header=0`` denotes the first line of data 
    rather than the first line of the file.
    
    .. note::
      Used only for ASCII tables, ignored by other types of tables.
    
  col_names : sequence of str
    Sequence of column labels to apply. If the file contains a header row, 
    then you should explicitly pass ``header=0`` to override the column names. 
    Duplicates in this list are not allowed.
    
    .. note::
      Used only for ASCII tables, ignored by other types of tables.
      
  dtype : 'pandas' or 'astropy'
    The data type returned by this function. 
    
    - If ``'pandas'`` (default) is specified, the table will be loaded as a 
      pandas dataframe
    - If ``'astropy'`` is specified, the table will be loaded as a astropy table

  Notes
  -----
  The Transportable Database Aggregate Table (TDAT) type is a data structure 
  created by NASA's Heasarc project and a very simple parser was implemented
  in this function due to lack of support in packages like pandas and astropy. 
  For more information, see [#TDAT]_

  Returns
  -------
  pd.DataFrame
    The table as a pandas dataframe

  Raises
  ------
  ValueError
    Raises an error if the file extension can not be detected
    
  References
  ----------
  .. [#TDAT] Transportable Database Aggregate Table (TDAT) Format.
      `<https://heasarc.gsfc.nasa.gov/docs/software/dbdocs/tdat.html>`_
  """
  is_file_like = False
  if isinstance(path, str):
    if path.startswith('http://') or path.startswith('https://') or path.startswith('ftp://'):
      fmt = fmt or path.split('.')[-1]
    else:
      path = Path(path)
      fmt = fmt or path.suffix
  if isinstance(path, Path):
    path = Path(path)
    fmt = fmt or path.suffix
  elif isinstance(path, pd.DataFrame):
    if columns is None:
      return path
    else:
      return path[columns].copy()
  elif isinstance(path, Table):
    df = path.to_pandas()
    if columns:
      df = df[columns]
    return df
  elif isinstance(path, (RawIOBase, BufferedIOBase, TextIOBase)):
    is_file_like = True
  
  if fmt.startswith('.'):
    fmt = fmt[1:]

  if fmt in ('fit', 'fits', 'fz'):
    with fits.open(path) as hdul:
      table_data = hdul[1].data
      table = Table(data=table_data)
      if columns is None:
        columns = [name for name in table.colnames if len(table[name].shape) <= 1]
      table = table[columns]
    if dtype == 'astropy':
      return table
    if dtype == 'pandas':
      return table.to_pandas()
  elif fmt in ('dat', 'tsv'):
    optional_params = {}
    if col_names:
      optional_params = {'names': col_names}
    df = pd.read_csv(
      path, 
      sep=r'\s+',
      usecols=columns, 
      low_memory=low_memory,
      comment=comment,
      na_values=na_values,
      keep_default_na=keep_default_na,
      na_filter=na_filter,
      header=header,
      **optional_params,
    )
    if dtype == 'pandas':
      return df
    if dtype == 'astropy':
      return Table.from_pandas(df)
  elif fmt == 'csv':
    optional_params = {}
    if col_names:
      optional_params = {'names': col_names}
    df = pd.read_csv(
      path, 
      usecols=columns, 
      low_memory=low_memory,
      comment=comment,
      na_values=na_values,
      keep_default_na=keep_default_na,
      na_filter=na_filter,
      header=header,
      **optional_params,
    )
    if dtype == 'pandas':
      return df
    if dtype == 'astropy':
      return Table.from_pandas(df)
  elif fmt == 'parquet':
    df = pd.read_parquet(
      path, 
      columns=columns,
    )
    if dtype == 'pandas':
      return df
    if dtype == 'astropy':
      return Table.from_pandas(df)
  elif fmt == 'feather':
    df = pd.read_feather(
      path, 
      columns=columns,
    )
    if dtype == 'pandas':
      return df
    if dtype == 'astropy':
      return Table.from_pandas(df)
  elif fmt == 'tdat':
    if is_file_like: raise ValueError('Not implemented for file-like objects')
    path = Path(path)
    content = path.read_text()
    header = re.findall(r'line\[1\] = (.*)', content)[0].replace(' ', '|')
    data = content.split('<DATA>\n')[-1].split('<END>')[0].replace('|\n', '\n')
    tb = header + '\n' + data
    df = pd.read_csv(
      StringIO(tb), 
      sep='|', 
      usecols=columns, 
      low_memory=low_memory
    )
    if dtype == 'pandas':
      return df
    if dtype == 'astropy':
      return Table.from_pandas(df)
  elif fmt in ('vo', 'vot', 'votable', 'xml'):
    result = votable.parse_single_table(path)
    table = result.to_table(use_names_over_ids=True)
    # table = result.get_first_table().to_table(use_names_over_ids=True)
    if columns:
      table = table[columns]
    if dtype == 'astropy':
      return table
    if dtype == 'pandas':
      return table.to_pandas

  raise ValueError(
    'Can not infer the load function for this table based on suffix. '
    'Please, use a specific loader.'
  )


def write_table(data: TableLike, path: PathOrFile, fmt: str | None = None):
  """
  Write a table to a file in various formats.

  Parameters
  ----------
  data : TableLike
    The table data to write. This can be an `astropy.table.Table` or a `pandas.DataFrame`.
  path : PathOrFile
    The path or file where the table will be written. This can be a string or a `pathlib.Path` object.
  fmt : str, optional
    The format in which to write the table. If not provided, the format is inferred from the file extension.

  Notes
  -----
  The function supports multiple output formats based on the file extension 
  or the ``fmt`` parameter:
  
  - ``'fits'``, ``'fit'``: writes to a FITS file.
  - ``'csv'``: writes to a CSV file.
  - ``'parquet'``: writes to a Parquet file.
  - ``'dat'``: writes to a space-separated text file.
  - ``'tsv'``: writes to a tab-separated text file.
  - ``'html'``: writes to an HTML file.
  - ``'feather'``: writes to a Feather file.
  - ``'vo'``, ``'vot'``, ``'votable'``, ``'xml'``: writes to a VOTable XML file.

  Raises
  ------
  ValueError
    If the provided format is not supported.

  Examples
  --------
  >>> from astropy.table import Table
  >>> import pandas as pd
  >>> data = Table({'col1': [1, 2], 'col2': [3, 4]})
  >>> write_table(data, 'output.fits')
  >>> df = pd.DataFrame({'col1': [1, 2], 'col2': [3, 4]})
  >>> write_table(df, 'output.csv')
  """
  if isinstance(path, str):
    fmt = fmt or Path(path).suffix
    Path(path).parent.mkdir(parents=True, exist_ok=True)
  elif isinstance(path, Path):
    fmt = fmt or path.suffix
    path.parent.mkdir(parents=True, exist_ok=True)
    path = str(path.absolute())
  
  if isinstance(data, Table):
    data = data.to_pandas()
  
  if fmt.startswith('.'):
    fmt = fmt[1:]
  
  if fmt in ('fit', 'fits'):
    Table.from_pandas(data).write(path, overwrite=True, format='fits')
  elif fmt == 'csv':
    data.to_csv(path, index=False)
  elif fmt == 'parquet':
    data.to_parquet(path, index=False)
  elif fmt == 'dat':
    data.to_csv(path, index=False, sep=' ')
  elif fmt == 'tsv':
    data.to_csv(path, index=False, sep='\t')
  elif fmt == 'html':
    data.to_html(path, index=False)
  elif fmt == 'feather':
    data.to_feather(path, index=False)
  elif fmt in ('vo', 'vot', 'votable', 'xml'):
    t = Table.from_pandas(data)
    votable.writeto(t, path)


def concat_tables(
  tables: Sequence[TableLike | PathOrFile],
  **kwargs
) -> pd.DataFrame:
  """
  Concatenate tables into a single one. This function concatenate over the
  table rows and is usefull to concatenate tables with same columns, although 
  there is no error in concatenating tables with non-existent columns in other.
  If a table does not have a certain column, the values will be filled with 
  ``NaN`` values. This function does not attempt to apply any type of 
  duplicate row removal.

  Parameters
  ----------
  tables : Sequence[TableLike  |  PathOrFile]
    A sequence of tables to be concatenated. This parameter accepts a
    table-like object (pandas dataframe, astropy table), a path to a file
    represented as a string or pathlib.Path object, or a file object
    (BinaryIO, StringIO, file-descriptor, etc).
  kwargs : Any
    Arguments that will be passed to `~astromodule.io.read_table` function

  Returns
  -------
  pd.DataFrame
    A dataframe of the concatenated table
  """
  dfs = [read_table(df, **kwargs) for df in tables]
  dfs = [df for df in dfs if isinstance(df, pd.DataFrame) and not df.empty]
  if len(dfs) > 0:
    return pd.concat(dfs)
  return pd.DataFrame()


def parallel_function_executor(
  func: Callable,
  params: Sequence[Dict[str, Any]] = [],
  workers: int = 2,
  callback: Callable = None,
  progress: bool = True,
  ignore_error: bool = True,
  unit: str = 'it'
):
  """
  Execute a function in parallel using multiple threads.

  Parameters
  ----------
  func : :term:`python:callable`
    The function to be executed in parallel.
  params : sequence of dict, optional
    A list of dictionaries containing the keyword arguments to be passed to 
    ``func`` for each call. Defaults to an empty list.
  workers : int, optional
    The number of worker threads to use. Defaults to 2.
  callback : :term:`python:callable`, optional
    A function to be called with the result of each ``func`` call. 
    Defaults to `None`.
  progress : bool, optional
    If `True`, displays a progress bar. Defaults to `True`.
  ignore_error : bool, optional
    If `True`, ignores exceptions raised by ``func`` calls. 
    Defaults to `True`.
  unit : str, optional
    The unit label to use in the progress bar. Defaults to ``'it'``.

  Raises
  ------
  Exception
    If ``ignore_error`` is `False`, raises the exception encountered during 
    the ``func`` execution.

  Notes
  -----
  The ``callback`` function will be called with the result of each successful 
  ``func`` execution. If ``ignore_error`` is False and an exception occurs 
  during the execution of ``func``, the exception will be raised and the 
  remaining tasks will not be executed.

  Examples
  --------
  >>> def func(x, y):
  ...     return x + y
  ...
  >>> params = [{'x': 1, 'y': 2}, {'x': 3, 'y': 4}]
  >>> parallel_function_executor(func, params, workers=2, callback=print)
  3
  7
  """
  with concurrent.futures.ThreadPoolExecutor(max_workers=workers) as executor:
    if progress:
      pbar = tqdm
    else:
      pbar = lambda *args, **kwargs: args[0]
      
    callback = callback or (lambda *args, **kwargs: None)
    
    futures = []
    for i in range(len(params)):
      futures.append(executor.submit(func, **params[i]))

    for future in pbar(
      concurrent.futures.as_completed(futures),
      total=len(futures),
      unit=unit
    ):
      try:
        callback(future.result())
      except Exception as e:
        if not ignore_error:
          raise e


import re
from datetime import datetime, timedelta
from multiprocessing import Lock
from pathlib import Path
from typing import Sequence, Tuple

import numpy as np
import pandas as pd
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.units import Quantity, Unit

RA_REGEX = re.compile(r'^ra_?J?\d*$', re.IGNORECASE)
DEC_REGEX = re.compile(r'^dec_?J?\d*$', re.IGNORECASE)




def _match_regex_against_sequence(
  regex: re.Pattern, 
  columns: Sequence[str]
) -> Tuple[int, str] | None:
  for i, col in enumerate(columns):
    if regex.match(col):
      return i, col
  return None

  
def guess_coords_columns(
  df: pd.DataFrame,
  ra: str | None = None,
  dec: str | None = None,
) -> Tuple[str, str]:
  """
  Receives a pandas dataframe and try to guess the columns names used to
  identify the RA and DEC coordinates.

  Parameters
  ----------
  df : pd.DataFrame
    A pandas dataframe
  ra : str | None, optional
    The column name used to name the RA column. If a string is passed, this
    function will skip the RA guessing and will return the value of RA passed
    by this parameter. If the value is set to ``None``, this function will
    guess the RA column name using a pre-defined regular expression and will
    return the value of the first match found following the sequence of
    the columns.
  dec : str | None, optional
    The column name used to name the DEC column. If a string is passed, this
    function will skip the DEC guessing and will return the value of DEC passed
    by this parameter. If the value is set to ``None``, this function will
    guess the DEC column name using a pre-defined regular expression and will
    return the value of the first match found following the sequence of
    the columns.

  Returns
  -------
  Tuple[str, str]
    A tuple of RA and DEC columns guessed.

  Raises
  ------
  ValueError
    Raises a error if the RA or DEC columns cannot be found.
  """
  cols = df.columns.to_list()
  if ra is None:
    _, ra = _match_regex_against_sequence(RA_REGEX, cols)
  if dec is None:
    _, dec = _match_regex_against_sequence(DEC_REGEX, cols)
  if ra is None or dec is None:
    raise ValueError(
      "Can't guess RA or DEC columns, please, specify the columns names "
      "via `ra` and `dec` parameters"
    )
  return ra, dec


def iauname(
  ra: float | np.ndarray,
  dec: float | np.ndarray
) -> str | Sequence[str]:
  """
  Receives the angular position(s) of the object(s) and returns IAU2000 name(s)

  Parameters
  ----------
  ra: float or array of float
    The right ascension of the object(s).
  dec: float or array of float
    The declination of the object(s).

  Example
  --------
  >>> iauname(187.70592, 12.39112)
  'J123049.42+122328.03'

  Returns
  -------
  str or list of str
    The formated IAU name of the object(s)
  """
  coord = SkyCoord(ra=ra*u.degree, dec=dec*u.degree, frame='icrs')
  ra_str = coord.ra.to_string(unit=u.hourangle, sep='', precision=2, pad=True)
  dec_str = coord.dec.to_string(sep='', precision=1, alwayssign=True, pad=True)
  if isinstance(ra_str, np.ndarray):
    r = [f'J{_ra_str}{_dec_str}' for _ra_str, _dec_str in zip(ra_str, dec_str)]
  else:
    r = f'J{ra_str}{dec_str}'
  return r



def iauname_path(
  iaunames: str | Sequence[str] = None,
  ra: float | Sequence[float] = None,
  dec: float | Sequence[float] = None,
  prefix: str | Path = '',
  suffix: str = '',
  flat: bool = False,
  return_str: bool = False,
) -> Path | Sequence[Path]:
  """
  Calculate the nested path for a given iauname

  Parameters
  ----------
  iaunames: str, List[str], optional
    Object iauname. The iauname or RA and DEC must be passed, if ``iaunames`` is
    ``None``, this function computes the iauname using the ``ra`` and ``dec``
    parameters
  ra: float, List[float], optional
    Object RA, used only if ``iaunames`` is ``None``
  dec: float, List[float], optional
    Object DEC, used only if ``iaunames`` is ``None``
  prefix: str, Path, optional
    Path that will be prepended at the begin of all paths
  suffix: str, optional
    Suffix that will be appended at the end of all paths
  flat: bool, optional
    Create the flat path with all files inside a same parent folder. This is
    not recomended for big datasets
  return_str: bool, optional
    Cast all paths to string before returning

  Example
  -------
  iaunames_path('J123049.42+122328.03', '.png')
  Path('J123/J123049.42+122328.03.png')

  Returns
  -------
  Path, List[Path]
    The iauname path
  """
  if iaunames is None:
    iaunames = iauname(ra, dec)

  if flat:
    mapping = lambda x:  Path(prefix) / (x + suffix)
  else:
    mapping = lambda x: Path(prefix) / x[:4] / (x + suffix)

  prep_output = lambda x: str(x) if return_str else x

  if isinstance(iaunames, str):
    return prep_output(mapping(iaunames))
  else:
    return [prep_output(mapping(x)) for x in iaunames]
  
  
def sanitize_quantity(
  quantity: float | int | Sequence | Quantity, 
  unit: str | Unit, 
  convert: bool = False
) -> Quantity:
  if isinstance(unit, str):
    unit = Unit(unit)
  if isinstance(quantity, str):
    if convert:
      return Quantity(quantity).to(unit)
    return Quantity(quantity)
  if isinstance(quantity, Quantity):
    if convert:
      return quantity.to(unit)
    return quantity
  return quantity * unit


class SingletonMeta(type):
  """
  Thread-safe implementation of Singleton.
  """
  _instances = {}
  """The dict storing memoized instances"""

  _lock = Lock()
  """
  Lock object that will be used to synchronize threads during
  first access to the Singleton.
  """

  def __call__(cls, *args, **kwargs):
    """
    Possible changes to the value of the `__init__` argument do not affect
    the returned instance.
    """
    # When the program has just been launched. Since there's no
    # Singleton instance yet, multiple threads can simultaneously pass the
    # previous conditional and reach this point almost at the same time. The
    # first of them will acquire lock and will proceed further, while the
    # rest will wait here.
    with cls._lock:
      # The first thread to acquire the lock, reaches this conditional,
      # goes inside and creates the Singleton instance. Once it leaves the
      # lock block, a thread that might have been waiting for the lock
      # release may then enter this section. But since the Singleton field
      # is already initialized, the thread won't create a new object.
      if cls not in cls._instances:
        instance = super().__call__(*args, **kwargs)
        cls._instances[cls] = instance
    return cls._instances[cls]


class Timer:
  def __init__(self, start: bool = True):
    self.start_time = None
    self.end_time = None
    if start:
      self.start()

  def __repr__(self) -> str:
    return self.duration

  def start(self):
    self.start_time = datetime.now()

  def end(self) -> timedelta:
    self.end_time = datetime.now()
    return self.duration

  @property
  def duration(self) -> timedelta:
    if not self.end_time:
      end_time = datetime.now()
      duration = end_time - self.start_time
    else:
      duration = self.end_time - self.start_time
    return duration
  
  @property
  def duration_str(self) -> str:
    return self._format_time(self.duration)
  
  def __str__(self):
    return self.duration_str

  def _format_time(self, dt: timedelta) -> str:
    hours, remainder = divmod(dt.seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    return '{:02}:{:02}:{:02}'.format(int(hours), int(minutes), int(seconds))